<?php

declare(strict_types=1);

namespace Drupal\page_cache_vary\StackMiddleware;

use Drupal\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\CacheableResponseInterface;
use Drupal\Core\PageCache\RequestPolicyInterface;
use Drupal\Core\PageCache\ResponsePolicyInterface;
use Drupal\page_cache\StackMiddleware\PageCache;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Extend the core PageCache and add the management of the Vary headers.
 *
 * To do so, we need to know what are the variations for a given URL: we need to
 * have resolved at least one request for a given URL.
 * Per specification, every responses under the same URL needs to return the
 * same VARY header, regardless of the variations, so we only need to have the
 * VARY header stored once per URL.
 *
 * This class relies on the base PageCache but only ensure that the CID is
 * generated using the VARY header: when a request is not in cache, no CID is
 * generated yet, it's only generated when the first response (containing the
 * VARY headers) is received.
 * The VARY is then put in cache, and every subsequent requests to the same URL
 * will use the cached VARY header.
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Vary
 */
final class PageCacheVary extends PageCache {

  /**
   * Cache ID for the Vary metadata.
   *
   * @var string|null
   */
  protected ?string $vcid = NULL;

  /**
   * Constructs a PageCacheVary object.
   */
  public function __construct(
    HttpKernelInterface $httpKernel,
    CacheBackendInterface $cachePage,
    RequestPolicyInterface $pageCacheRequestPolicy,
    ResponsePolicyInterface $pageCacheResponsePolicy,
    protected ContainerInterface $container,
    protected CacheBackendInterface $cachePageMetadata,
    protected array $varyCacheContexts,
  ) {
    parent::__construct($httpKernel, $cachePage, $pageCacheRequestPolicy, $pageCacheResponsePolicy);
  }

  /**
   * Return FALSE if there is currently no Cache ID.
   *
   * It means that this URL has not yet been computed.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Current request.
   * @param bool $allow_invalid
   *   Allow invalid cache to be returned.
   *
   * @return false|Response
   *   False if there is no cache available, the response otherwise.
   */
  protected function get(Request $request, $allow_invalid = FALSE) {
    if (!$this->getCacheId($request)) {
      return FALSE;
    }

    return parent::get($request, $allow_invalid);
  }

  /**
   * Generate the correct vary cache entry when a response is available.
   *
   * Retrieve the vary headers from the cache contexts.
   * It's expected (by the norm) to have the same vary headers for every call on
   * the same resource, so we don't need to merge vary headers whatsoever.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   * @param \Symfony\Component\HttpFoundation\Response $response
   *   The response to save in the cache.
   * @param int $expire
   *   The expiry time, in seconds.
   * @param array $tags
   *   The cache tags of the response.
   */
  protected function set(Request $request, Response $response, $expire, array $tags) {

    // An URL needs to return the same Vary headers regardless of the value of
    // the varied headers.
    // @see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Vary
    // Since the response might change the vary headers, we need to re-process
    // them every time, which will update it for every possible variation of
    // this response.
    $varyHeaders = $this->getVaryHeadersFromResponse($response);

    $this->cachePageMetadata->set(parent::getCacheId($request), $varyHeaders);

    $this->vcid = $this->getVaryHeadersCid(
      $varyHeaders,
      $request,
    );

    $response->setVary($varyHeaders, FALSE);

    parent::set($request, $response, $expire, $tags);
  }

  /**
   * Compute the cache ID based on both the base cache ID and the variations.
   *
   * Since it's not possible to know the variations until the response is
   * completed, if there is no entries in the cachePageMetadata it means that no
   * responses have been computed.
   * The response needs to be computed for the given cache variation.
   *
   * Return NULL when a variation is expected, but the computation has not yet
   * been done.
   *
   * @return string|null
   *   The current cache ID if all the values used to compute it are available.
   *
   * @see \Drupal\page_cache\StackMiddleware\PageCache::getCacheId
   */
  protected function getCacheId(Request $request) {
    $urlCid = parent::getCacheId($request);

    if (!$this->vcid) {
      if ($cachedCidVaries = $this->cachePageMetadata->get($urlCid)) {
        $this->vcid = $this->getVaryHeadersCid($cachedCidVaries->data, $request);
      }
      else {
        return NULL;
      }
    }

    return parent::getCacheId($request) . ':' . $this->vcid;
  }

  /**
   * Retrieve the Vary Headers from the response.
   *
   * Uses the cache contexts of the given response to get the Vary Headers
   * from.
   *
   * @param \Symfony\Component\HttpFoundation\Response $response
   *   The request's computed response.
   *
   * @return array
   *   List of vary headers from the given response.
   */
  protected function getVaryHeadersFromResponse(Response $response): array {
    $varies = [];

    if ($response instanceof CacheableResponseInterface) {
      $contexts = $response->getCacheableMetadata()->getCacheContexts();

      foreach ($contexts as $context) {
        $contextName = $this->getContextIdFromContext($context);

        if (in_array($contextName, $this->varyCacheContexts)) {
          $varies += $this->container->get('cache_context.' . $contextName)->getVaryHeaders();
        }
      }

    }

    return $varies;
  }

  /**
   * Compute the CID of the vary headers.
   *
   * @param array $headers
   *   List of headers used for Vary.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return string
   *   Computation of all the vary headers from the request, in a CID sanitized
   *   string.
   */
  protected function getVaryHeadersCid(array $headers, Request $request): string {
    // Sort the vary headers to be sure we always get the correct cache entry.
    sort($headers);

    $headersValue = [];

    foreach ($headers as $header) {
      $headersValue[] = $header . '=' . base64_encode($request->headers->get($header, ''));
    }

    return implode(':', $headersValue);
  }

  /**
   * Get the clean context ID from the contexts of the response.
   *
   * Sometimes the context contains not only the context ID, but also some
   * parameters (user:authenticated).
   * To retrieve the service we need to clear the parameters.
   *
   * @param string $context
   *   The context of the request.
   *
   * @return string
   *   The "clean" context corresponding to the context ID. Example:
   *   user:authenticated will be transformed into user.
   */
  protected function getContextIdFromContext(string $context): string {
    if ($position = strpos($context, ':')) {
      return substr($context, 0, $position);
    }
    else {
      return $context;
    }
  }

}
