<?php

namespace Drupal\Tests\page_cache_vary\Functional;

use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\system\Functional\Cache\AssertPageCacheContextsAndTagsTrait;
use Drupal\page_cache_vary_test\Cache\Context\CountryCacheContext;

/**
 * Test the Cache Vary to ensure that it behave as expected.
 */
class PageCacheVaryTest extends BrowserTestBase {

  use AssertPageCacheContextsAndTagsTrait;

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'olivero';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'page_cache_vary',
    'page_cache_vary_test',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->enablePageCaching();
  }

  /**
   * Test the cache behavior respect variations.
   *
   * Test that a page without the country Vary cache context will hit the cache
   * on every call, regardless of the value of the x-country-code header.
   * Ensure that a page with the country Vary cache context varies depending on
   * the value of x-country-code, but the cache when the vary header has already
   * been stored in cache.
   */
  public function testPageCacheVary() {

    // Ensure that without cache, we get the correct response regardless of the
    // varied header.
    $this->drupalGet('/page-cache-vary-test/without-country');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('No country here');
    $this->assertNoCacheContext('country');
    $this->assertCacheMiss();

    $this->drupalGet('/page-cache-vary-test/without-country');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertCacheHit();

    $this->drupalGet(
          '/page-cache-vary-test/without-country', headers: [
            CountryCacheContext::COUNTRY_CODE_HEADER => 'ch',
          ]
      );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertNoCacheContext('country');
    $this->assertCacheHit();

    $this->drupalGet(
          '/page-cache-vary-test/without-country', headers: [
            CountryCacheContext::COUNTRY_CODE_HEADER => 'ch',
          ]
      );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertNoCacheContext('country');
    $this->assertCacheHit();

    // Ensure that with country context, the response vary depending on the
    // country.
    $this->assertCountryDisplayed('', FALSE);
    $this->assertCountryDisplayed('', TRUE);

    $this->assertCountryDisplayed('ch', FALSE);
    $this->assertCountryDisplayed('fr', FALSE);

    $this->assertCountryDisplayed('ch', TRUE);
    $this->assertCountryDisplayed('fr', TRUE);

    // Ensure that changing the context will update the cache.
    $config = \Drupal::configFactory()->getEditable('page_cache_vary_test.settings');
    $config->set('vary', FALSE)->save();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'ch',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('No country here');
    $this->assertNoCacheContext('country');
    $this->assertCacheMiss();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'fr',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('No country here');
    $this->assertNoCacheContext('country');
    $this->assertCacheHit();

    // Change the vary to clear the response cache. Expect the metatatags cache
    // to be updated.
    $config->set('vary', TRUE)->save();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'ch',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains("Country: Switzerland");
    $this->assertCacheContext('country');
    $this->assertCacheMiss();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'fr',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains("Country: France");
    $this->assertCacheContext('country');
    $this->assertCacheMiss();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'ch',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains("Country: Switzerland");
    $this->assertCacheContext('country');
    $this->assertCacheHit();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'fr',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains("Country: France");
    $this->assertCacheContext('country');
    $this->assertCacheHit();

    // The cache should be updated.
    $config->set('vary', FALSE)->save();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'ch',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('No country here');
    $this->assertNoCacheContext('country');
    $this->assertCacheMiss();

    $this->drupalGet(
      '/page-cache-vary-test/alternating-country', headers: [
        CountryCacheContext::COUNTRY_CODE_HEADER => 'fr',
      ]
    );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('No country here');
    $this->assertNoCacheContext('country');
    $this->assertCacheHit();
  }

  /**
   * Test that the country is correctly handled.
   *
   * Assert that the given country is correctly returned in the page and the
   * cache is correctly handled.
   *
   * @param string|null $countryCode
   *   Country code to test (either ch, fr or '').
   * @param bool $cacheHitExpected
   *   True if a cache hit is expected, false if a cache miss is expected.
   */
  protected function assertCountryDisplayed(string $countryCode, bool $cacheHitExpected) {

    $countryName = match($countryCode) {
      'ch' => 'Switzerland',
            'fr' => 'France',
            '' => 'Unknown',
    };

    // Headers are stored in the underline driver, we need to ensure it's reset
    // between every request.
    $this->getSession()->setRequestHeader(CountryCacheContext::COUNTRY_CODE_HEADER, '');
    $this->drupalGet(
          '/page-cache-vary-test/with-country', headers: [
            CountryCacheContext::COUNTRY_CODE_HEADER => $countryCode,
          ]
      );
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains("Country: {$countryName}");

    if ($cacheHitExpected) {
      $this->assertCacheHit();
    }
    else {
      $this->assertCacheMiss();
    }
  }

  /**
   * Asserts that the cache is hit by checking the response header.
   */
  protected function assertCacheHit() {
    $this->assertSession()->responseHeaderContains('X-Drupal-Cache', 'HIT');
  }

  /**
   * Asserts that the cache is missed by checking the response header.
   */
  protected function assertCacheMiss() {
    $this->assertSession()->responseHeaderContains('X-Drupal-Cache', 'MISS');
  }

}
